// File:  RdbStatsPercentiles.cc

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of rdbstats
//
// rdbstats is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rdbstats is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <sstream>
#include <algorithm>

#include <suplibxx/str.h>

#include "RdbStatsPercentiles.h"

// The user specified percentiles.  For example: --percentiles 12,34,45
RdbStatsPercentiles::RdbStatsPercentiles( RDB& irdbtable, const string& name,
					  const string& percent_list,
					  bool implicitquartile )
  throw( Exception ) try : RdbStats( irdbtable, name ), the_median( 0.0 ),
			   the_percentile_result( NULL ) {

    try {

#ifdef TRACEFCT
      TraceFct tf( "RdbStatsPercentiles::RdbStatsPercentiles( RDB&, "
		   "const string&, const string& ) throw( Exception )" );
#endif

      vector< string > tokens;
  
      suplib::tok( tokens, percent_list, "," );

      const int mysize( tokens.size( ) );
      for ( int ii = 0; ii < mysize; ii++ ) {
	string sval = string( "_p" ) + tokens[ ii ];
	double dval = suplib::str2d( tokens[ ii ].c_str( ) );
	the_percentile.push_back( pair< string, double >( sval, dval ) );
      }

      for ( int ii = 0; ii < mysize; ii++ ) {
	if ( the_percentile[ ii ].second <= 0.0 ||
	     the_percentile[ ii ].second > 100.0 ) {
	  ostringstream ost;
	  ost << "The percentiles option (" << percent_list << ") must be "
	    "within [ 0.0, 100.0 ]\n";
	  throw Exception( ost.str( ) );
	}
      }

      if ( implicitquartile ) {
	the_percentile.push_back( pair< string, double >( "_fq", 25.0 ) );
	the_percentile.push_back( pair< string, double >( "_lq", 75.0 ) );
      }

    } catch ( Exception& e ) {

      throw;

    } catch ( exception& e ) {

      throw Exception( e.what( ) );

    }

  } catch ( Exception& E ) {

    // exception thrown by RdbStats class is caught here
    throw;
    
  } catch ( exception& e ) {
    
    // exception thrown by RdbStats class is caught here
    throw Exception( e.what( ) );

  }


// --quartiles was entered at the command line.
RdbStatsPercentiles::RdbStatsPercentiles( RDB& irdbtable, const string& name )
  throw( Exception ) try : RdbStats( irdbtable, name ), the_median( 0.0 ),
			   the_percentile_result( NULL ) {

    try {

#ifdef TRACEFCT
      TraceFct tf( "RdbStatsPercentiles::RdbStatsPercentiles( RDB&, "
		   "const string&, const string& ) throw( Exception )" );
#endif

      the_percentile.push_back( pair< string, double >( "_fq", 25.0 ) );
      the_percentile.push_back( pair< string, double >( "_lq", 75.0 ) );

    } catch ( Exception& e ) {

      throw;

    } catch ( exception& e ) {

      throw Exception( e.what( ) );

    }

  } catch ( Exception& E ) {
    
    // exception thrown by RdbStats class is caught here
    throw;
    
  } catch ( exception& e ) {
    
    // exception thrown by RdbStats class is caught here
    throw Exception( e.what( ) );

  }


double RdbStatsPercentiles::calculate_median(  ) throw( ) {

#ifdef TRACEFCT
    TraceFct tf( "double RdbStatsPercentiles::calculate_mdeian( ) throw( )" );
#endif

  vector< double >::iterator begin_ptr = data.begin( );

  size_t num_size = data.size( );
  size_t num_size_div = num_size / 2;

  nth_element( begin_ptr, begin_ptr + num_size_div, data.end( ) );

  if ( num_size & 0001 ) {

    // num_size is odd:

    return data[ num_size_div ];

  } else {

    // num_size is even:

    double tmp = data[ num_size_div ];

    //
    // element beyond num_size_div is guarranteed
    // to be > then  element at data[ num_size_div ].
    //
    nth_element( begin_ptr, begin_ptr + num_size_div - 1,
		 begin_ptr + num_size_div );
    tmp += data[ num_size_div - 1 ];

    return 0.5 * tmp;

  }

}

double RdbStatsPercentiles::calculate_percentile( const double percentile )
  throw( ) {

#ifdef TRACEFCT
    TraceFct tf( "double RdbStatsPercentiles::calculate_percentile( int ) "
		 "throw( )" );
#endif

  vector< double >::iterator begin_ptr = data.begin( );

  size_t n        = data.size( );
  double f        = percentile / 100.0;
  int i           = ( ( n - 1.0 ) * f );
  double delta    = ( n - 1.0 ) * f - i; 

  nth_element( begin_ptr, begin_ptr + i + 1, data.end( ) );
  double data_i_1 = data[ i + 1];

  nth_element( begin_ptr, begin_ptr + i, data.end( ) );
  double data_i   = data[ i ];

  double quantile = ( 1.0 - delta ) * data_i + delta * data_i_1;

  /*
  cerr << "n:         " << n        << endl
       << "f:         " << f        << endl
       << "i:         " << i        << endl
       << "delta:     " << delta    << endl
       << "data_i:    " << data_i   << endl
       << "data_i_1:  " << data_i_1 << endl
       << "quantile:  " << quantile << endl << endl;
  */
  return quantile;

}

int RdbStatsPercentiles::calculate_statistics( ) throw( ) {

#ifdef TRACEFCT
    TraceFct tf( "double RdbStatsPercentiles::calculate_statistics( ) "
		 "throw( )" );
#endif

  if ( 0 == this->RdbStats::calculate_statistics( ) )
    return 0;

  the_median = calculate_median( );

  for ( size_t ii = 0; ii < the_percentile.size( ); ii++ )
    the_percentile_result[ ii ] =
      calculate_percentile( the_percentile[ ii ].second );

  return num_n;

}

void RdbStatsPercentiles::init( ) throw( ) {

#ifdef TRACEFCT
  TraceFct tf( "void RdbStatsPercentiles:init( ) throw( )" );
#endif

  this->RdbStats::init( );

  // Since we are, possibly, starting a new group so wipe out data.
  data.clear( );

}

void RdbStatsPercentiles::normalize_results( const double norm ) throw( ) {

  this->RdbStats::normalize_results( norm );
  for ( size_t ii = 0; ii < the_percentile.size( ); ii++ )
    the_percentile_result[ ii ] /= norm;

}

void RdbStatsPercentiles::set_output_columns( RDB& ordbtable )
  throw( Exception ) {

  try {

#ifdef TRACEFCT
    TraceFct tf( "double RdbStatsPercentiles::set_output_columns( RDB& ) "
		 "throw( Exception )" );
#endif
    
    this->RdbStats::set_output_columns( ordbtable );

    const int mysize( the_percentile.size( ) );
    the_percentile_result = new double[ mysize ];

    const char* column_name_ptr = column_name.c_str( );

    char str[ 256 ];
    sprintf( str, "%s_median", column_name_ptr );
    ordbtable.setColumn( str, "N" );
    RDBColumn* tmp = ordbtable.getColumn( str );
    tmp->mapData( &the_median, 1 );

    for ( int ii = 0; ii < mysize; ii++ ) {
      
      ostringstream ost;
      ost << column_name_ptr << the_percentile[ ii ].first;
      ordbtable.setColumn( ost.str( ), "N" );
      RDBColumn* ptr = ordbtable.getColumn( ost.str( ) );
      ptr->mapData( the_percentile_result + ii , 1 );

    }

  } catch ( RDBErr& rdbe ) {

    throw Exception( rdbe );

  } catch( Exception& E ) {

    throw;

  } catch ( exception& e ) { 

    throw Exception( e.what( ) );

  }
      
}

void RdbStatsPercentiles::update_statistics( ) throw( Exception ) {

  try {

#ifdef TRACEFCT
    TraceFct tf( "double RdbStatsPercentiles::update_statistics( ) "
		 "throw( Exception )" );
#endif

    this->RdbStats::update_statistics( );

    data.push_back( get_value( ) );

  } catch( Exception& e ) {

    throw;

  }

}
