// File:  Row.cc

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of rdbstats
//
// rdbstats is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rdbstats is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <functional>
#include <sstream>

#include <algorithm>

#include <Exception/Exception.h>
#include <suplib/range.h>

#include "Row.h"

inline bool mycmp( const pair< long, long >& a, const pair< long, long >& b ) {
  return a.first < b.first;
}

Row::Row( const vector< string >& row_options ) throw ( Exception ) {

  vector< string >::const_iterator iter = row_options.begin( );
  vector< string >::const_iterator the_end = row_options.end( );

  if ( iter == the_end ) {
    // If the user did not request a specific row then use the entire file.
    start_end.push_back( pair< long, long > ( 1, LONG_MAX ) );
    return;
  }

  for ( ; iter != the_end; ++iter ) {

    RangeLList* rll = NULL;

    RangeOpts opt =
      (RangeOpts) ( Range_SORT | Range_MERGE | Range_INCOMPLETE );
    long where, start = 1, end = LONG_MAX;

    RangeErr myerr;
    if ( (myerr=rangel_parse( &rll, (char*) iter->c_str( ), opt,
			      start, end, &where )) ) {
      rangel_del( rll );
      ostringstream ost;
      ost << "Row::Row( ) : Unable to parse '" << *iter << "'\n";
      throw Exception( ost.str( ) );
    }

    for ( size_t ii = 0; ii < rll->n ; ii++ )
      start_end.push_back( pair< long, long > ( rll->range[ ii ].start,
						rll->range[ ii ].end ) );

    sort( start_end.begin( ), start_end.end( ), ptr_fun( mycmp ) );

    rangel_del( rll );
    
  }

}

void Row::print( ostream& os ) const {

  vector< pair< long, long > > ::const_iterator iter = start_end.begin( );
  vector< pair< long, long > >::const_iterator end = start_end.end( );
  for ( ; iter != end; ++iter ) {
    os << '[' << iter->first << ',' << iter->second << ']';
    if ( iter + 1 != end )
      os << '\n';
  }

}

#ifdef Row_Testme

int main( int argc,  char* argv[] ) {

  try {

    vector< string > options;
    for ( int ii = 1; ii < argc; ++ii )
      options.push_back( argv[ ii ] );

    Row row( options );

    cout << row << '\n';

    return 0;

  } catch( Exception& E ) {

    cerr << E << '\n';
    return 1;

  } catch( exception& e ) {

    cerr << e.what( ) << 'n';
    return 1;

  } catch( ... ) {

    cerr << "unknown exception caught\n";
    return 1;

  }

}

/*
 CC -DRow_Testme MyRow.cc -I/proj/axaf/simul/include /proj/axaf/simul/lib/sun4u-SunOS-5/libException.a /proj/axaf/simul/lib/sun4u-SunOS-5/libsuplib.so
*/

#endif
