// File:  SelectedCols.cc

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of rdbstats
//
// rdbstats is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rdbstats is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <iterator>

#include "SelectedCols.h"

SelectedCols::SelectedCols( const vector< string >& cols, RDB& rdbtable,
			    clo::parser& clo ) throw( Exception ) {

#ifdef TRACEFCT
  TraceFct tf( "SelectedCols( const vector< string >&, RDB&, "
	       "clo::parser& ) throw( Exception )" );
#endif

  try {

    const clo::options& options = clo.get_options();

    for ( int ii = 0; ii < cols.size( ); ii++ )
      if ( true == is_column_numeric( cols[ ii ], rdbtable, options ) )
	selected_cols.push_back( cols[ ii ] );

  } catch ( Exception& e ) {

    throw;

  }

}

SelectedCols::SelectedCols( RDB& rdbtable, clo::parser& clo ) 
 throw( Exception ) {

#ifdef TRACEFCT
  TraceFct tf( "SelectedCols( RDB&, clo::parser& ) throw( Exception )" );
#endif

  try {

    const clo::options& options = clo.get_options();

    // --group a -- group b
    vector< string > groupies;
    vector< string >::const_iterator current_group( options.group.begin( ) ),
      end_group( options.group.end( ) );
    for ( ; current_group != end_group; ++current_group ) {
      // --group a,b,c may have been entered, so must parse entry.
      suplib::tok( groupies, *current_group, "," );
    }

    //
    // Loop through all rdb column and select the numeric columns only.
    //
    size_t num = rdbtable.nColumns( );

    for ( size_t ii = 0; ii < num; ii++ ) {

      string name = rdbtable.getColumn( ii )->getName( );

      if ( true == is_column_numeric( name, rdbtable, options ) &&
	   false == is_groupie( name, groupies ) )
	selected_cols.push_back( name );

    }

  } catch ( RDBErr& rdbe ) {

    throw Exception( rdbe );

  } catch ( Exception& e ) {

    throw;

  }

}

bool SelectedCols::is_column_numeric( const string& colname,
				      RDB& rdbtable ) throw( Exception ) {

#ifdef TRACEFCT
  TraceFct tf( "SelectedCols::is_column_numeric( const string&, "
	       "RDB& ) throw( Exception )" );
#endif

  try {

    RDBColumn::Type mytype = rdbtable.getColumn( colname )->getType( );

    if ( RDBColumn::NUMERIC == mytype )
      return true;
    else
      return false;

  } catch ( RDBErr& rdbe ) {

    throw Exception( rdbe );

  } catch ( Exception& e ) {

    throw;

  }

}

bool SelectedCols::is_column_numeric( const string& colname,
				      RDB& rdbtable,
				      const clo::options& options )
  throw ( Exception ) {

#ifdef TRACEFCT
  TraceFct tf( "SelectedCols::is_column_numeric( const string&, "rr
	       "RDB&, const clo::options& ) throw( Exception )" );
#endif

  try {

    bool is_numeric( false );

    RDBColumn::Type mytype = rdbtable.getColumn( colname )->getType( );

    if ( RDBColumn::NUMERIC == mytype )
      is_numeric = true;

    vector< string >::const_iterator
      current_override( options.override.begin( ) ),
      end_override( options.override.end( ) );

    for ( ; current_override != end_override; ++current_override ) {

      string str( *current_override );

      //
      // The user has requested that a column is to change type.
      //
      vector< string > container;

      suplib::tok( container, str, "," );

      if ( 0 == colname.compare( container[ 0 ] ) ) {

	switch( container.size( ) ) {

	case 1:

	  //
	  // The definition is unspecified so set to
	  // the opposite of the current definition.
	  //
	  toggle_column_definition( colname, rdbtable );
	  return is_column_numeric( colname, rdbtable );

	case 2:

	  //
	  // Set the column definition to the user specified value.
	  //
	  rdbtable.getColumn( colname )->setDef( container[ 1 ] );
	  return is_column_numeric( colname, rdbtable );

	default:

	  throw Exception( "Too many options for column " + colname );

	}

      }

    }

    return is_numeric;

  } catch ( RDBErr& rdbe ) {

    throw Exception( rdbe );
      
  } catch ( Exception& e ) {

    throw;

  } 

}

bool SelectedCols::is_groupie( const string& name,
			       const vector<string>& groupies ) {

  vector< string >::const_iterator current_group( groupies.begin( ) ),
      end_group( groupies.end( ) );
  for ( ; current_group != end_group; ++current_group )
    if ( 0 == name.compare( *current_group ) )
      return true;

  return false;

}

void SelectedCols::print( ostream& os ) const {

#ifdef TRACEFCT
  TraceFct tf( "SelectedCols::print( ostream& ) const " );
#endif

  copy( selected_cols.begin( ), selected_cols.end( ),
	ostream_iterator< string >( os, " " ) );

}

void SelectedCols::toggle_column_definition( const string& colname,
					     RDB& rdbtable )
  const throw ( Exception ) {

#ifdef TRACEFCT
  TraceFct tf( "SelectedCols::toggle_column_definition( const string&, "
	       "RDB& ) throw( Exception )" );
#endif

  try {

    RDBColumn::Type mytype = rdbtable.getColumn( colname )->getType( );
    if ( RDBColumn::NUMERIC == mytype )
      rdbtable.getColumn( colname )->setDef( "S" );
    else
      rdbtable.getColumn( colname )->setDef( "N" );

  } catch ( RDBErr& rdbe ) {

    throw Exception( rdbe );

  }

}

