// File:  rdbstats.cc

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of rdbstats
//
// rdbstats is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rdbstats is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <iostream>
#include <cstdlib>

#include <rdbxx/RDB.h>
#include <Exception/Exception.h>

#include "Options.h"
#include "usage.h"
#include "RdbStats.h"
#include "RdbStatsAve.h"
#include "RdbStatsPercentiles.h"
#include "RdbStatsPercentilesAve.h"
#include "RdbStatsPercentilesMed.h"
#include "Row.h"
#include "SelectedCols.h"
#include "StatsResult.h"
#include "config.h"

using namespace std;

/**
   Use Rows or groups to select to calculate the statistics.
 */
template<class Type>
void rdbstats( RDB& irdbtable, Type& rdb_stats_columns,
	       const vector< pair< long, long > >& range, RDB& ordbtable, bool grouped )
  throw( Exception ) {

#ifdef TRACEFCT
  TraceFct tf( "rdbstats( RDB&, Type&, const vector< pair<long, long> >&,"
	       "RDB&, bool ) throw( Exception )" );
#endif

  try {

    StatsResult< Type >* stats_result_ptr =
      new StatsResult< Type >( rdb_stats_columns, ordbtable );

    size_t mysize = range.size( );
    int status = RDB::REOF;

    for ( int ii = 0, line_number = 1; ii < mysize; ++ii, ++line_number ) {

      // Skip the un-requested rows.
      // At the end of this loop, line_number == range[ ii ].first
      for ( ; line_number < range[ ii ].first; ++line_number ) {
	if ( RDB::REOF == (status = irdbtable.read( )) ) {

	  stats_result_ptr->calculate_statistics( );
	  delete stats_result_ptr;
	  return;

	}
	//	ordbtable.autoIdx( false );
      }

      // loop through the requested range of rows.
      for ( ; line_number <= range[ ii ].second; ++line_number ) {

	if ( RDB::REOF == (status = irdbtable.read( )) ) {

	  if ( ! grouped ) 
	    stats_result_ptr->calculate_statistics( ); 

	  delete stats_result_ptr;

	  return;

	} 

	stats_result_ptr->update_statistics( );

	if ( RDB::REOG & status ) 

	  stats_result_ptr->calculate_statistics( );

	//      	ordbtable.autoIdx( false );

      }

    }

    delete stats_result_ptr;

  } catch( Exception& e ) {

    cerr << e << '\n';
    throw;

  }

}

/**
 * To decide what kind of statistics to be calculated!
 */
void rdbstats( clo::parser& clo ) throw( Exception ) {

#ifdef TRACEFCT
  TraceFct tf( "rdbstats( clo::parser& ) throw( Exception& e )" );
#endif

  try {

    /*
     * get the struct of options from the parser class so that you can
     * access the option values.
     */
    const clo::options& options = clo.get_options();
    const std::vector<std::string>& the_files = clo.get_non_options();

    // The input stream shall be the standard in
    RDB *irdbtable = ( options.input == "stdin" ) ? ( new RDB( &cin ) ) : ( new RDB( options.input ) );

    // The output stream shall be the standard out
    //    ostream* os = &cout;
    //    os->precision( 15 );
    RDB *ordbtable = ( options.output == "stdout" ) ? ( new RDB( &cout ) ) : ( new RDB( options.output ) );
    ordbtable->setComment( *irdbtable );

    SelectedCols* selected_cols = NULL;
    if ( options.all || 0 == the_files.size( ) ) {
      selected_cols = new SelectedCols( *irdbtable, clo );
    } else {
      selected_cols = new SelectedCols( the_files, *irdbtable, clo );
    }
    // cout << *selected_cols << '\n';

    const vector< string >& col_name = selected_cols->get_selected_cols( );
    // cout << "# col_name.size( ) = " << col_name.size( ) << '\n';
    if ( 0 == col_name.size( ) )
      return;

    // Set the group.
    bool grouped = false;
    vector< string >::const_iterator current_group( options.group.begin( ) ),
      end_group( options.group.end( ) );
    for ( ; current_group != end_group; ++current_group ) {
      // --group a,b,c may have been entered, so must parse entry.
      vector< string > groupies;
      suplib::tok( groupies, *current_group, "," );
      for ( unsigned int ii = 0; ii < groupies.size( ); ii++ ) {
	ordbtable->setColumn( irdbtable->getColumn( groupies[ ii ] ) );
	irdbtable->setGroup( groupies[ ii ] );
	grouped = true;
      }
    }

    Row row( options.rows );
    // cout << '#' << row << '\n';

    if ( options.percentiles != "" ) {

      string percentiles( options.percentiles );

      //
      // This allows the possibility that the user entered :   --p 23,45 -q
      //
      bool iq = options.quartiles;

      switch( options.normalize ) {
      case clo::normalize_ave:
	{
	  vector< RdbStatsPercentilesAve > rdb_stats_columns;

	  for ( unsigned int ii = 0; ii < col_name.size( ); ii++ )
	    rdb_stats_columns.push_back( RdbStatsPercentilesAve( *irdbtable,
								 col_name[ii],
								 percentiles,
								 iq ) );
	  rdbstats( *irdbtable, rdb_stats_columns, row.get_range( ),
		    *ordbtable, grouped );
	}
	break;
      case clo::normalize_med:
	{
	  vector< RdbStatsPercentilesMed > rdb_stats_columns;
	  for ( unsigned int ii = 0; ii < col_name.size( ); ii++ )
	    rdb_stats_columns.push_back( RdbStatsPercentilesMed( *irdbtable,
								 col_name[ii],
								 percentiles,
								 iq ) );
	  rdbstats( *irdbtable, rdb_stats_columns, row.get_range( ),
		    *ordbtable, grouped );
	}
	break;
      case clo::normalize_none:
	{
	  vector< RdbStatsPercentiles > rdb_stats_columns;
	  for ( int ii = 0; ii < col_name.size( ); ii++ )
	    rdb_stats_columns.push_back( RdbStatsPercentiles( *irdbtable,
							      col_name[ ii ],
							      percentiles,
							      iq ) );
	  rdbstats( *irdbtable, rdb_stats_columns, row.get_range( ),
		    *ordbtable, grouped );
	}
	break;
      }

    } else if ( options.quartiles ) {

      switch( options.normalize ) {
      case clo::normalize_ave:
	{
	  vector< RdbStatsPercentilesAve > rdb_stats_columns;
	  for ( unsigned int ii = 0; ii < col_name.size( ); ii++ )
	    rdb_stats_columns.push_back( RdbStatsPercentilesAve( *irdbtable,
								 col_name[ ii ]
								 ) );
	  rdbstats( *irdbtable, rdb_stats_columns, row.get_range( ),
		    *ordbtable, grouped );
	}
	break;
      case clo::normalize_med:
	{
	  vector< RdbStatsPercentilesMed > rdb_stats_columns;
	  for ( unsigned int ii = 0; ii < col_name.size( ); ii++ )
	    rdb_stats_columns.push_back( RdbStatsPercentilesMed( *irdbtable,
								 col_name[ ii ]
								 ) );
	  rdbstats( *irdbtable, rdb_stats_columns, row.get_range( ),
		    *ordbtable, grouped );
	}
	break;
      case clo::normalize_none:
	{
	  vector< RdbStatsPercentiles > rdb_stats_columns;
	  for ( unsigned int ii = 0; ii < col_name.size( ); ii++ )
	    rdb_stats_columns.push_back( RdbStatsPercentiles( *irdbtable,
							      col_name[ ii ] )
					 );
	  rdbstats( *irdbtable, rdb_stats_columns, row.get_range( ),
		    *ordbtable, grouped );
	}
	break;
      }

    } else {

      switch( options.normalize ) {
      case clo::normalize_ave:
	{
	  vector< RdbStatsAve > rdb_stats_columns;

	  for ( unsigned int ii = 0; ii < col_name.size( ); ii++ )
	    rdb_stats_columns.push_back( RdbStatsAve( *irdbtable,
						      col_name[ ii ] ) );
	  rdbstats( *irdbtable, rdb_stats_columns, row.get_range( ),
		    *ordbtable, grouped );
	}
	break;
      case clo::normalize_med:
	{
	  // implicit --quartiles case
	  vector< RdbStatsPercentilesMed > rdb_stats_columns;
	  for ( unsigned int ii = 0; ii < col_name.size( ); ii++ )
	    rdb_stats_columns.push_back( RdbStatsPercentilesMed( *irdbtable,
								 col_name[ ii ]
								 ) );
	  rdbstats( *irdbtable, rdb_stats_columns, row.get_range( ),
		    *ordbtable, grouped );
	}
	break;
      case clo::normalize_none:
	{
	  vector< RdbStats > rdb_stats_columns;
	  for ( unsigned int ii = 0; ii < col_name.size( ); ii++ )
	    rdb_stats_columns.push_back( RdbStats( *irdbtable, col_name[ ii ]
						   ) );
	  rdbstats( *irdbtable, rdb_stats_columns, row.get_range( ),
		    *ordbtable, grouped );
	}
	break;
      }

    }

    delete selected_cols;
    delete ordbtable;
    delete irdbtable;

  } catch( RDBErr& rdbe ) {

    cerr << rdbe << '\n';
    throw Exception( rdbe );

  } catch( Exception& e ) {

    cerr << e << '\n';
    throw;

  } catch( exception& ee ) {

    cerr << ee.what( ) << '\n';
    throw Exception( ee.what( ) );

  } catch( ... ) {

    throw Exception( "Caught unknown exception, got to love C++!" );

  }

}

int main( int argc, char* argv[] ) {

#ifdef TRACEFCT
  TraceFct tf( argv[0], 1, -1 );
#endif

  try {

    clo::parser		clo;
    clo.parse( argc, argv );

    if ( 0 != clo.get_options( ).group.size( ) &&
	 0 != clo.get_options( ).rows.size( ) )
      throw Exception(  "The options '--group'and '--rows' "
			"cannot be used in conjunction\n" );

    rdbstats( clo );

    return EXIT_SUCCESS;

  } catch ( clo::autoexcept& cloe ) {

    switch ( cloe.get_autothrow_id() ) {
    case clo::autothrow_help:
      std::cout << "Usage: " << argv[0] << " [options]\n";
      std::cout << cloe.what();
      return EXIT_SUCCESS;
    case clo::autothrow_usage:
      usage( );
      return EXIT_SUCCESS;
    case clo::autothrow_version:
      std::cout << PACKAGE_NAME << '\t' << PACKAGE_VERSION << endl;
      return EXIT_SUCCESS;
    }

  } catch( Exception& E ) {

    cerr << E << endl;
    return EXIT_FAILURE;

  } catch( std::exception &e ) {

    std::cerr << e.what() << "\n";
    return EXIT_FAILURE;

  } catch ( ... ) {

    cerr << "Unknown exception caught, got to love C++\n";
    return EXIT_FAILURE;

  }

  return 0;

}
